/*  GtkExText	
*  Copyright (C) 1999 by:
*Mikael Hermansson <mikeh@bahnhof.se>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details. 
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include "gdk/gdkkeysyms.h"
#include "gtkextext.h"
#include "gtkextext_static.h"

#define TEXT_BORDER_ROOM 2
#define MIN_TEXT_WIDTH_LINES     20
#define MIN_TEXT_HEIGHT_LINES    10
#define MAX_LINE_LEN 300
#define TEXT_BUFFERT_DEFAULT_SIZE 0x8000
#define GAP_DEFAULT_SIZE TEXT_BUFFERT_DEFAULT_SIZE 
#define SCROLL_TIME 100
#define DEFAULT_TAB_STOP  5
#define DEBUG
enum {
  ARG_0,
  ARG_HADJUSTMENT,
  ARG_VADJUSTMENT,
};

enum {
	PROPERTY_INSERT,
	PROPERTY_REMOVE,
	UNDO_CHANGED,
	PROPERTY_MARK,
	LINE_INSERT,
	LINE_REMOVE,
	LAST_SIGNAL
};

static guint extext_signals[LAST_SIGNAL]={ 0 };


typedef void (*GtkExTextFunction)(GtkEditable *text, gint time);
static const GtkExTextFunction control_keys[26] =
{
  NULL,    /* a */
  NULL,   /* b */
  (GtkExTextFunction)gtk_editable_copy_clipboard,        /* c */
  NULL,  /* d */
  NULL,          /* e */
  NULL,    /* f */
  NULL,                                                /* g */
  NULL, /* h */
  NULL,                                                /* i */
  NULL,                                                /* j */
  (GtkExTextFunction)gtk_extext_delete_to_line_end,        /* k */
  NULL,                                                /* l */
  NULL,                                                /* m */
  NULL,            /* n */
  NULL,                                                /* o */
  NULL,        /* p */
  NULL,                                                /* q */
  NULL,                                                /* r */
  NULL,                                                /* s */
  NULL,                                                /* t */
  (GtkExTextFunction)gtk_extext_delete_line,               /* u */
  (GtkExTextFunction)gtk_editable_paste_clipboard,       /* v */
  (GtkExTextFunction)gtk_extext_delete_backward_word,      /* w */
  (GtkExTextFunction)gtk_editable_cut_clipboard,         /* x */
  NULL,                                                /* y */
  NULL,                                                /* z */
};

static const GtkExTextFunction alt_keys[26] =
{
  NULL,                                                /* a */
  NULL,       /* b */
  NULL,	/* c */                      
	NULL,                          /* d */
  NULL,                                           /* e */
  NULL,         /* f */
  NULL,                                           /* g */
  NULL,                                           /* h */
  NULL,                                           /* i */
  NULL,                                           /* j */
  NULL,                                           /* k */
  NULL,                                           /* l */
  NULL,                                           /* m */
  NULL,                                           /* n */
  NULL,                                           /* o */
  NULL,                                           /* p */
  NULL,                                           /* q */
  NULL,                                           /* r */
  NULL,                                           /* s */
  NULL,                                           /* t */
  NULL,                                           /* u */
  NULL,                                           /* v */
  NULL,                                           /* w */
  NULL,                                           /* x */
  NULL,                                           /* y */
  NULL,                                           /* z */
};


/*****************************************************************************/
/*                          hmm lets start coding :-)                            */
/*****************************************************************************/

gint get_next_return(gchar *text,gint len)
{
	gint index;
	for(index=0;index<len;index++)	{
		if(text[index]=='\n')
			return index;
	}
	return -1;
}
gint get_tab_count(gchar *txt,gint len)
{
	gint tabs;
	gint i;
	tabs=i=0;
	if(len==-1)
		len=strlen(txt);

	while(i<len)
	{	
		if(txt[i]=='\t')
			tabs++;

		i++;
	};
	return tabs;
}
gint get_return_count(gchar *txt,gint len)
{
	gint ret;
	gint i;
	ret=i=0;
	while(i<len)
	{	
		if(txt[i]=='\n')
			ret++;

		i++;
	};
	return ret;
}
gint is_value_ok(gint pos,gint start,gint end)
{
	if(pos>=start && pos<end)
		return 0;
	if(pos<end)
		return -1;

	return 1;
}

gint get_real_text_height(GdkFont *font)
{
	gint lbearing;
	gint rbearing;
	gint width;
	gint ascent_;
	gint descent_;
	gdk_string_extents(font,
                   	" `~!@#$%^&*()-_=+\\|[]{};:\"\'<,>.?/1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
                   	&lbearing,
                   	&rbearing,
                   	&width,
                   	&ascent_,
                   	&descent_);

	return (ascent_+descent_)* 1.20;
}

gint get_text_descent(GdkFont *font)
{
	gint lbearing;
	gint rbearing;
	gint width;
	gint ascent_;
	gint descent_;
	gdk_string_extents(font,
                   	" `~!@#$%^&*()-_=+\\|[]{};:\"\'<,>.?/1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
                   	&lbearing,
                   	&rbearing,
                   	&width,
                   	&ascent_,
                   	&descent_);

	return descent_;
}

gint get_real_text_width(GdkFont *font)
{
	gint lbearing;
	gint rbearing;
	gint width;
	gint ascent_;
	gint descent_;
	gdk_string_extents(font,
                   	" `~!@#$%^&*()-_=+\\|[]{};:\"\'<,>.?/1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
                   	&lbearing,
                   	&rbearing,
                   	&width,
                   	&ascent_,
                   	&descent_);

	return width;
}

/* check for previous word and */
/* updates start and endpos */
void gtk_extext_get_previous_word(GtkExText *text,gint *startpos, gint *endpos)
{
	guchar key;
	if(!startpos || !endpos)
		return ;
	

	*endpos=*startpos;
	*startpos=*startpos+1;
	while(*startpos>0)
	{
		*startpos=*startpos-1;
		key=gtk_extext_get_char_at_pos(text,*startpos);
		if(!isalnum(key))	{		
			*startpos=*startpos+1;
			break;
		}
	}

	*endpos=*startpos-1;
	while(*endpos<text->length)
	{
		*endpos=*endpos+1;
		key=gtk_extext_get_char_at_pos(text,*endpos);
		if(!isalnum(key))	{		
			break;
		}
	}
		
}
void gap_to(GtkExText *text,gint pos)
{
	gint i;
	if(pos==text->part1len)
		return;
	
	if(pos<text->part1len)	{
		for(i=0;i<text->part1len-pos;i++)
			text->text[text->part1len+text->gap_len - i -1]=text->text[text->part1len-i-1];
	}
	else	{
		for(i=0;i<pos-text->part1len;i++)
			text->text[text->part1len+i]=text->text[text->part1len+text->gap_len+i];
	}
	text->part1len=pos;
	text->part2text=text->text+text->gap_len;
}
void make_room_for(GtkExText *text,gint len)
{
	gint newsize;
	gint count;

	if(text->gap_len<=len)
	{
		gap_to(text,text->length);

		count=len/text->size;
		if(count*text->size<len)
			count++;
		
		newsize=text->gap_len + text->length + (text->size * count);
		text->text=g_realloc(text->text,newsize);

		text->gap_len=newsize - text->length;
		text->part2text=text->text + text->gap_len;
	}
}

gint text_set_pos(GtkExText *text,gint pos)
{
	LineCache* line;
	gboolean forward;
	gint lindex=0;
	LineCache* oldline;

	oldline=text->line_pos;
	forward=FALSE;
	if(pos>=GTK_EDITABLE(text)->current_pos)
		forward=TRUE;
		
	
	if(pos>text->length)
		GTK_EDITABLE(text)->current_pos=text->length;
	else if(pos==-1)
		GTK_EDITABLE(text)->current_pos=0;
	else
		GTK_EDITABLE(text)->current_pos=pos;

	line=line_get_by_char_pos(text,pos,&lindex);
	if(line!=oldline)	{
		line_set_by_ptr(text,line,lindex,forward);
	}	
	line_set_cursor_pos(text,GTK_EDITABLE(text)->current_pos 
							- line_get_start_pos(text,line));


	return GTK_EDITABLE(text)->current_pos;
}

gint text_insert( GtkExText *text,gint pos,gchar *txt,gint len,gboolean removethis)
{
	gint start;
	gint end;
	gint linecurpos;
	LineCache  *line;
	gint lindex;
	gboolean noreturn;
	gint diff;
	gint retcount;
	gint signal_return_value;

  g_return_val_if_fail (text != NULL,FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (text),FALSE);

	signal_return_value=FALSE;
	noreturn=TRUE;
	if(!len)
		return text->length;

	if(pos != GTK_EDITABLE(text)->current_pos)
		pos=text_set_pos(text,pos);

	make_room_for(text,len);
	gap_to(text,pos);

	line=text->line_pos;
	lindex=text->line_pos_index;
	retcount=get_return_count(txt,len);
	end=get_next_return(txt,len);

	start=end+1;
	if(end!=-1)	{
		gboolean retvalue;
		retvalue=FALSE;
		noreturn=FALSE;
			/* linecurpos has chars before this insert at currline */

		linecurpos=text->line_cursor_index;
			/* diff is all chars after return in curline */ 
			/*	this chars is moved to newline */

		diff=line_get_length(text,text->line_pos) 
						-  text->line_cursor_index;

		line_change(text,line,linecurpos+end+1);

		line=line_insert(text,line,retcount,&lindex);

		end=get_next_return(&txt[start],len-start);
		if(end==-1)
		{	line=line_change(text,line->next,diff);	lindex++;	}
		else
		{
			while(end!=-1)
			{
				retcount--;
				line=line_change(text,line->next,end+1);
				lindex++;

				start+=end+1;
				end=get_next_return(&txt[start],len-start);


				linecurpos=0; /* we have inserted a new line and has no chars yet*/
			};
			start+=end+1;
			line=line_change(text,line->next,diff+(len - start));
		}
	}
	
	memcpy(text->text+text->part1len,txt,len);
	
	text->part1len+=len;
	text->length+=len;
	text->gap_len-=len;
	text->part2text=text->text+text->gap_len;

	extext_undo_insert(text,EXTEXT_UNDO_INSERT,pos, pos+len);


	if(noreturn ) {	/* nope no new returns only new chars added to curline */
		linecurpos=line_get_length(text,text->line_pos);
		line_change(text,line,linecurpos + len); 
		text_set_pos(text,text->part1len);
	}
	else{
		text_set_pos(text,text->part1len);
	}


	gtk_signal_emit(GTK_OBJECT(text),extext_signals[PROPERTY_INSERT],
								gtk_extext_property_nearest_backward(text,pos,NULL),pos,pos+len,&signal_return_value);
	
	if(!signal_return_value) {
		gtk_extext_property_move_all(text,pos,len,NULL); 
	}

	/* MUST RETURN pos so it passes correct to gtk_editable signal callbaks */
	/* HMMMM is this really correct ????????? */

	return pos+len;
}
gint text_remove(GtkExText *text,gint pos,gint len)
{
	GtkExTextProperty *startprop,*next;
	LineCache* line;
	gint i;
	gint oldlinelen;
	gint lindex;
	gint linelen;
	gint signal_return_value;

	startprop=next=NULL;
	signal_return_value=FALSE;
	
	if(len==0)
		return GTK_EDITABLE(text)->current_pos;

	if(pos==-1)
		return GTK_EDITABLE(text)->current_pos;
	if(len==-1)
		return GTK_EDITABLE(text)->current_pos;

	if(pos>text->length)
		pos= text->length;
	
	if(pos>=GTK_EDITABLE(text)->selection_start_pos)
		GTK_EDITABLE(text)->selection_start_pos=-1;
	if(len+pos<=GTK_EDITABLE(text)->selection_end_pos)
		GTK_EDITABLE(text)->selection_end_pos=-1;

	extext_undo_insert(text,EXTEXT_UNDO_REMOVE,pos, pos+len);

	lindex=0;
	line=line_get_by_char_pos(text,pos,&lindex);
	if(!line)	
	{ g_warning("[text_remove] line don't exist?????\n");	return -1;	}

	linelen=line->len;
	for(i=pos;i<len+pos;i++)	{
		if(gtk_extext_get_char_at_pos(text,i)=='\n')	{	
			/* always remove line AFTER Current line*/
			/* this because we should never remove FIRST line */
			/* ask for all characters before RETURN char */	
			line=line->next;
			lindex++;
			oldlinelen=line_get_length(text,line)-1;
			line=line_remove(text,line,&lindex);
			line->len+=oldlinelen;
			linelen=line->len;
		}
		else	{
			linelen--;
			line->len=linelen;
		}
	}

	gap_to(text,pos);
	text->length -=len;
	text->gap_len +=len;
	text->part2text=text->text + text->gap_len;

	text_set_pos(text,pos);

	startprop=gtk_extext_property_nearest_backward(text,pos,NULL);
	gtk_signal_emit(GTK_OBJECT(text),extext_signals[PROPERTY_REMOVE],
					startprop,pos,pos+len,&signal_return_value);
	if(!signal_return_value) 	{
		/* remove propertys between start and endpos */
		next=gtk_extext_property_remove_all(text,pos,pos+len,startprop);
		gtk_extext_property_move_all(text,pos,-(len-pos),next);
	}


	return GTK_EDITABLE(text)->current_pos;
}
guchar gtk_extext_get_char_at_pos(GtkExText *text,gint pos)
{
	guchar txt;

	if(pos>=text->part1len)
		txt=text->part2text[pos];
	else
		txt=text->text[pos];

	return txt;
}
guchar * text_get_text_with_expanded_tabs(GtkExText *text,gint start, gint end  ,gint *tabindex )
{
	guchar *txt,*newtxt;
	gint len;
	gint tabcount;
	gint i,n;

	i=n=0;

	len=end-start;

	txt=text_get_text(text,start,end);
	if(!txt)
		return txt;


	tabcount=get_tab_count(txt,len);	
	if(tabcount)	{
		newtxt=g_malloc0(len + (tabcount*5) + 2);
		for(i=0;i<len;i++)	{
			if(txt [i]=='\t')	{
				strcpy(&newtxt[n],"     ");
				n+=5;
			}
			else
				newtxt[n++]=txt[i];
		};
		g_free(txt);
	}
	else
		newtxt=txt;

	*tabindex+=tabcount;

	return newtxt;
}
guchar* text_get_text(GtkExText *text,gint startpos,gint endpos)
{
	guchar *txt;
	gint i;
	gint len;

	if(endpos<startpos)	{
		g_warning("Startpos is more than endpos[text_get_text]\n");
		return NULL;
	}
	if(startpos>text->length)	{
		g_warning("Startpos out of range[text_get_text]\n");
		return NULL;
	}
	if(endpos>text->length)
		endpos=text->length;
	
	if(startpos==endpos)
		return NULL;

	len=endpos-startpos;
	txt=g_new0(guchar,len+1);
	for(i=0;i<len;i++)
	{
		txt[i]=gtk_extext_get_char_at_pos(text,startpos++);
	}
	txt[i]=0;
	
	return txt;
}

/* the main handler to set cursor at pos */
/* called by text_set_pos */
/* and goto_line  */
 
static gint line_set_cursor_pos(GtkExText *text,gint cursorpos)
{
	LineCache* oldline; 
	gint len,oldcol;	
	GtkExTextProperty *prop;

	prop=text->property_current;
	oldcol=text->line_cursor_index;
	oldline=text->line_pos;
	len=line_get_length(text,text->line_pos);
	if(cursorpos < 0 ) {
		line_set_by_ptr(text,text->line_pos->prev,text->line_pos_index-1,FALSE);

		if(oldline!=text->line_pos)
			text->line_cursor_index=line_get_length(text,text->line_pos)-1;
		else
			text->line_cursor_index=0;
	}
	else if(cursorpos >= len)	{
		line_set_by_ptr(text,text->line_pos->next,text->line_pos_index+1,TRUE);
		
		if(oldline==text->line_pos)
			text->line_cursor_index=len;
		else
			text->line_cursor_index=0;
	}
	else
 		text->line_cursor_index=cursorpos;

	GTK_EDITABLE(text)->current_pos=line_get_start_pos(text,text->line_pos) 
								+ text->line_cursor_index;


	text->property_current=gtk_extext_property_get_at_pos(text,
											GTK_EDITABLE(text)->current_pos,prop);

	if(prop!=text->property_current)
		gtk_signal_emit_by_name(GTK_OBJECT(text),"property_mark");

	if(oldcol!=text->line_cursor_index)
		gtk_signal_emit_by_name(GTK_OBJECT(text),"move-to-column",
						text->line_cursor_index);

	return text->line_cursor_index;
}
static LineCache* line_change(GtkExText *text,LineCache *cur,gint len)
{
	if(!text )
		return 0;

	if(!cur)
	{
		g_warning("line_change LineCache *NULL???\n");
		return NULL;
	}


	if(cur && len!=-1)	{
		cur->len=len;		
	}		
	return cur;
}
static LineCache* line_insert(GtkExText *text,LineCache *pos,gint count,gint *lindex)
{
	/* INSERT AFTER */
	
	gint i;
	LineCache *cur;
	LineCache *newc;
	LineCache *after;
	
	after=NULL;
	cur=pos;
	if(cur)	
		after=cur->next;

	for(i=0;i<count;i++)	{
		newc=g_malloc0(sizeof(LineCache));
		if(!text->line_start) /* no lines at all */
			text->line_start=newc;

		newc->prev=cur;	
		if(cur)
			cur->next=newc;		

		cur=newc;

   	 gtk_signal_emit(GTK_OBJECT(text),extext_signals[LINE_INSERT],*lindex+i+1);
	}
	cur->next=after;
	if(after)
		after->prev=cur;

	text->line_count+=count;

	if(!cur->next)
		text->line_end=cur;

	/* return the previously line */
	return pos;
}
gint line_get_num(GtkExText *text,LineCache *cur,gboolean forward)
{
	LineCache *start;
	gint i=0;
	
	if(!text->line_count)
		return 0;
	
	#ifdef DEBUG
	if(!cur)
	{	g_warning("LineCache *NULL in line_get_num???\n");	return -1;	}
	#endif
	if(!cur->next)
		return text->line_count;

	if(forward)	{
		start=text->line_pos;
		i=text->line_pos_index;
		while(start)	{
			if(cur==start)
				return i;
		
			start=start->next;

			i++;
		}
	} 
	else	{
		start=text->line_pos;
		i=text->line_pos_index;
		while(start)	{
			if(cur==start)
				return i;
		
			i--;
			start=start->prev;
		}
	}
	g_warning("lineptr %X don't exist\n",(gint)cur);

	return -1;
}
/* this call should only be used by the Intern API */

static LineCache * line_get_ptr(GtkExText * text,gint *line)
{
	gint i;
	LineCache *cache;

	/* if first line or line -1 */
	if(*line<=0 || !text->line_count){	
		*line=0; 

		return text->line_start; 
	}
	if(*line>=text->line_count)
	{	
		*line=text->line_count; 
		return text->line_end; 
	}

	if(*line>=text->line_pos_index)
	{
		i=text->line_pos_index;
		cache=text->line_pos;
		while(cache)
		{
			if(i==*line) {
				*line=i;

				return cache;
			}
			cache=cache->next;
			i++;
		};
	}		
	else	{
		i=text->line_pos_index;
		cache=text->line_pos;
		while(cache)
		{
			if(i==*line) {
				*line=i;

				return cache	;
			}
			cache=cache->prev;
			i--;
		};
	}
	

	g_warning("LINE DONT EXIST percent error?????"
			" %d last is %d  line_pos_index is %d\n",
				*line,text->line_count,text->line_pos_index);

	return NULL;
}

static gint line_get_length(GtkExText *text, LineCache *cur)
{
	if(!cur)
		return 0;

	return cur->len;

}
static LineCache* line_remove(GtkExText *text,LineCache *cur,gint *lindex)
{
	LineCache *prev=NULL;
	LineCache *remove=NULL;

	if(!cur || cur==text->line_start)
	{	*lindex=0;	cur=NULL; prev=text->line_start;	}
	else if(cur==text->line_end) {
		text->line_end=text->line_end->prev;
	}

	if(cur)	{	
   	 gtk_signal_emit(GTK_OBJECT(text),extext_signals[LINE_REMOVE],*lindex);

		prev=cur->prev;
		*lindex=*lindex-1;
		remove=cur;
		if(prev)
			prev->next=cur->next;
		if(cur->next)
			cur->next->prev=prev;
		else
			cur->next=NULL;

		if(remove)
			g_free(remove);

		text->line_count--;
	}

	return prev;
}
LineCache* line_get_by_char_pos(GtkExText *text,gint textoffset,gint *lineindex)
{
	gint offset;
	LineCache *cur;	

	cur=text->line_start;
	offset=0;
	if(textoffset>=text->length)
	{	*lineindex=text->line_count; return text->line_end;	 }

	*lineindex=0;
	while(cur) {
		offset+=cur->len;
		if(offset>textoffset)
			return cur;
		
		*lineindex=*lineindex+1;
		cur=cur->next;
	}
	return NULL;
}
gint line_get_end_pos(GtkExText *text,LineCache *pos)
{
	return line_get_end_pos_by_offset(text,pos,0);
}
gint line_get_end_pos_by_offset(GtkExText *text,LineCache *pos,gint offset)
{
	LineCache *cur;
	gint i;
		
	if(!pos)
		return 0;

	offset=0;
	i=0;
	cur=text->line_start;
	while(cur)	{
		offset+=cur->len;
		if(pos==cur)	{
			return offset;
		}
		i++;
		cur=cur->next;
	};
	return text->length;
}
gint line_get_start_pos(GtkExText *text,LineCache *line)
{
	if(!line){
		g_warning("Calling line_get_start_pos with LineCache * = NULL ????\n");
		return 0;
	}
	if(!line->prev)
		return 0;

	return line_get_end_pos(text,line->prev);
}
/* line_set_ptr set line by pointer and index */
/* if you dont know thje line index set it to -1  */
/* gboolean forward==TRUE searches forward in list index list */
/* search from end in list */ 

LineCache* line_set_by_ptr(GtkExText *text,LineCache *pos,gint index,gboolean forward)
{
	gint oldline;

	oldline=text->line_pos_index;	
	if(!pos && text->line_pos) /* npo dont set it */
	{	text->line_pos=text->line_pos; index=text->line_pos_index;	}
	else if(!pos)
	{	text->line_pos=text->line_start; index =0; }
	else 
		text->line_pos=pos;
			

	if(index==-1)
		text->line_pos_index=line_get_num(text,text->line_pos,forward);
	else
		text->line_pos_index=index;

	if(oldline!=text->line_pos_index)
		gtk_signal_emit_by_name(GTK_OBJECT(text),"move-to-row",	
						text->line_pos_index);

	return text->line_pos;	
}

LineCache* line_set(GtkExText *text,gint pos)
{
	gint lindex=pos;
	if(pos>text->line_count)	{
		line_set_by_ptr(text,text->line_end,text->line_count,FALSE); 
	}
	else if(pos<0)	{
		line_set_by_ptr(text,text->line_start,0,TRUE); 
	}
	else	{ 
		LineCache *lptr;
		lptr=line_get_ptr(text,&lindex);
		line_set_by_ptr(text,lptr,lindex,FALSE);
	}	

	return text->line_pos;	
}
/*****************************************************************************/
/*                           StyleCache declarations                             */
/*****************************************************************************/

void cb_style_destroy(gpointer freeme,gpointer data)
{
	g_free(freeme);
}
void cbh_style_destroy(gpointer key,gpointer freeme,gpointer data)
{
	g_free(freeme);
}
void property_class_destroy(GtkExText *text)
{
	GtkExTextProperty *next,*cur;

	g_hash_table_foreach(text->hash_styles,(GHFunc)cbh_style_destroy,NULL);
	g_hash_table_destroy(text->hash_styles);

	cur=text->property_start;
	while(cur)
	{
		next=cur->next;
		g_free(cur);
		cur=next;
	};
	
}

GtkExTextStyle *gtk_extext_style_insert(GtkExText *text,gchar *key,GdkFont *font,GdkColor *fg,GdkColor *bg)
{
	GtkExTextStyle *style,*def;
	
	def=gtk_extext_style_get(text,"Default");
	if((!font || !bg || !fg  ) && !def)	{
		g_warning("Widget must be realized before you can insert propertys\n");
		return NULL;
	}
	style=gtk_extext_style_get(text,key);	
	if(!style)	{
		style=g_malloc0(sizeof(GtkExTextStyle));
		g_hash_table_insert(text->hash_styles,(gpointer *)key,style);
	}
	if(!font)
		style->font=def->font;
	else
		style->font=font;

	if(!fg)
		style->fg_color=def->fg_color;
	else	{
		gdk_color_alloc(gdk_colormap_get_system(),fg);
		style->fg_color=*fg;
	}

	if(!bg )
		style->bg_color=def->bg_color;
	else	{
		gdk_color_alloc(gdk_colormap_get_system(),bg);
		style->bg_color=*bg;
	}

	return style;
}

void gtk_extext_style_remove(GtkExText *text, gchar *key)
{
	GtkExTextStyle *style;
	if(!strcasecmp(key,"Default") && !strcasecmp(key,"Selected") 
				&& !strcasecmp(key,"Prelight")){		
			g_warning("you cannot remove default/selected/Prelight style\n"
					"use gtk_extext_style_insert to change the values\n");	

			return  ;
	}

	style=gtk_extext_style_get(text,key);

	g_free(style);
	g_hash_table_remove(text->hash_styles,key);
}

GtkExTextStyle * gtk_extext_style_get(GtkExText *text, gchar *key)
{
	GtkExTextStyle *style;

	style=(gpointer)g_hash_table_lookup(text->hash_styles,key);
	return style;
}
/* we dont use  g_list because we need direct acces to ->next */

GtkExTextProperty * property_remove_linked_list(GtkExText *text,GtkExTextProperty *cur)
{
	GtkExTextProperty *next;
	if(!cur )
		return NULL;

	if(!cur->prev && !cur->next) /* last property */
	{	 text->property_start=text->property_end=NULL; return NULL; }

	next=cur->next;
	if(!next) {	/* LAST property */
		text->property_end=cur->prev;
		text->property_end->next=NULL;

	}
	else     /* somewhere in list */
	{
		if(cur->prev)
			cur->prev->next=next;

		next->prev=cur->prev;
	}

	if(!text->property_end || !text->property_end->prev)
		text->property_start=text->property_end;

	if(text->property_start==cur)
		text->property_start=next;

	return next;
}

void property_insert_linked_list(GtkExText *text,GtkExTextProperty *cur,GtkExTextProperty *next)
{
	if(!cur)
		return;

	cur->prev=NULL;
	cur->next=NULL;
	/* INSERT BEFORE */
	if(next)
	{
		cur->prev=next->prev;
		next->prev=cur;
		cur->next=next;
		if(cur->prev)
			cur->prev->next=cur;
	}

	if(!cur->next) { /* there is no after cur property_end */
		if(text->property_end)
			text->property_end->next=cur;
			
		cur->prev=text->property_end;
		text->property_end=cur;
	}

	if(!cur->prev) /* cur is the first in list */
		text->property_start=cur;
}
GtkExTextProperty* gtk_extext_property_remove_all(GtkExText *text,gint startpos,gint endpos,GtkExTextProperty *startprop)
{
	gint i;
	GtkExTextProperty *next;
	GtkExTextProperty *prev;
	next=NULL;

	prev= startprop ? startprop : text->property_start;

	i=startpos;
	do {
		prev=gtk_extext_property_get_at_pos(text,i,prev);
		if(prev)	{
			next=property_remove_linked_list(text,prev);
			g_free(prev);
			prev=next;
		}	
		
		i++;
	}while(i<endpos);

	return next ? next : gtk_extext_property_nearest_forward(text,endpos,prev) ;
}
GtkExTextProperty* gtk_extext_property_move_all(GtkExText *text,gint pos,gint diff,GtkExTextProperty *prop)
{
	GtkExTextProperty *next;

	next=NULL;
	if(!prop)
		prop=text->property_start;

	while(prop)
	{
		if(prop->startpos>=pos)
		{
			if(!next)	{
				next=prop;
				if(!diff)
					return next;
			}
			prop->startpos+=diff;
			prop->endpos+=diff;
		}
		prop=prop->next;
	}
	return next;
}

GtkExTextProperty * gtk_extext_property_get_current(GtkExText *text)
{
	g_return_val_if_fail (GTK_IS_EXTEXT (text), NULL);
	
	return text->property_current;
}

GtkExTextProperty * gtk_extext_property_insert(GtkExText *text,gchar *key,gint startpos,gint endpos, gpointer data,gboolean append)
{
	GtkExTextProperty *prop,*next;
	GtkExTextStyle *style;
	
	next=NULL;
	if(key && *key)
	{
		style=gtk_extext_style_get(text,key);
		if(!style && *key)	{	
			g_warning("No style to the property\n");
			return NULL; 
		}
	}	
	prop=g_malloc(sizeof(GtkExTextProperty));
	prop->startpos=0;
	prop->endpos=0;
	prop->data=data;	/* for future implements */
	prop->next=NULL;
	prop->prev=NULL;

	if(key && *key)
		prop->style_key=g_strdup(key);		
	else		
		prop->style_key=NULL;


	if(append) /* NO we dont update em. We only get the next prop to pass to linked_list */
		next=gtk_extext_property_move_all(text,startpos,0,NULL); 
	else	{
		next=gtk_extext_property_remove_all(text,startpos,endpos,NULL);
	}

	property_insert_linked_list(text,prop,next);

	prop->startpos=startpos;
	prop->endpos=endpos;

	return prop;
}

GtkExTextProperty* gtk_extext_property_remove(GtkExText *text,GtkExTextProperty *prop)
{
	GtkExTextProperty *next;

	if(!prop)
		return text->property_end;

	next=property_remove_linked_list(text,prop);
	g_free(prop);

	return next;     /* ALWAYS return NEXT property */
}

/* ALL propertys should be in sorted order */
/*****************************************************/
GtkExTextProperty *gtk_extext_property_get_at_pos(GtkExText *text,gint pos,GtkExTextProperty *oldprop)
{
	GtkExTextProperty *cur,*curstart;

	if(!text->property_start)
		return NULL;
		
	curstart = cur = oldprop  ? oldprop : text->property_start;	
	while(cur)
	{
		gint val;
		val=is_value_ok(pos,cur->startpos,cur->endpos);
		/* if return 0, pos is in property interval :-) */
		if(!val)
			return cur;

		/* if value returns more it probadly because property -> startpos */
		/*  is after pos */
		if(val<0)
			return NULL;

		cur=cur->next;
	};

	/* we started from curstart so we should start before curstart */
	cur=curstart->prev;

	/*	search backward if not found forward */	
	/* until val is less than cur->endpos */
	while(cur)
	{
		gint val;
		val=is_value_ok(pos,cur->startpos,cur->endpos);
		/* if val return 0 it is found */
		if(!val)
			return cur;

		if(val>0)
			return NULL;

		cur=cur->prev;
	};

	return NULL;
}
GtkExTextProperty *property_get_nearest_at_pos(GtkExText *text,gint pos,gboolean forward,GtkExTextProperty *cur)
{
	if(!cur)
		cur=text->property_start;	
	
	if(!cur)
		return NULL;

	/* if between */
	if(pos>=cur->startpos && pos<cur->endpos)
		return cur;

	while(cur && !forward)
	{
		if(cur->startpos>pos)
			return cur->prev;

		cur=cur->next;
	};
	while(cur && forward)
	{
		if(cur->startpos>=pos)
			return cur;

		cur=cur->next;
	};

	return NULL;
}

GtkExTextProperty *gtk_extext_property_nearest_forward(GtkExText *text,gint pos,GtkExTextProperty *cur)
{
	return property_get_nearest_at_pos(text,pos,TRUE,cur);	
}
GtkExTextProperty *gtk_extext_property_nearest_backward(GtkExText *text,gint pos,GtkExTextProperty *cur)
{
	return property_get_nearest_at_pos(text,pos,FALSE,cur);	
}


/*****************************************************************************/
/*                       Static       Undo code                              */
/*****************************************************************************/

void extext_undo_foreach(gpointer data,gpointer user_data)
{
	g_free(data);
}

void extext_undo_insert(GtkExText *text,gint type, gint start, gint end)
{
	UndoBuffert *undo;

	if(text->undo_flag || !text->undo_max) /* we inserting text from undo buffert */
		return ;			/* so we should NOT insert it in buffert again */

	undo=g_malloc0(sizeof(UndoBuffert));

	undo->type=type;
	undo->startpos=start;
	undo->endpos=end;
	if(undo->type==EXTEXT_UNDO_REMOVE)
		undo->buffert=text_get_text(text,start,end);
	
     if(!text->undo_buffert)  /* first undo ? */
   	 gtk_signal_emit(GTK_OBJECT(text),extext_signals[UNDO_CHANGED]);

	text->undo_buffert=g_slist_append(text->undo_buffert,undo);
	text->undo_count++;
	text->undo_last++;
}

/*****************************************************************************/
/*                                 widget api                                   */
/*****************************************************************************/


static void
gtk_extext_set_arg (GtkObject        *object,
		  GtkArg           *arg,
		  guint             arg_id)
{
  GtkExText *text;
  
  text = GTK_EXTEXT (object);
  
  switch (arg_id)
    {
    case ARG_HADJUSTMENT:
      gtk_extext_set_adjustments (text,
				GTK_VALUE_POINTER (*arg),
				text->vadj);
      break;
    case ARG_VADJUSTMENT:
      gtk_extext_set_adjustments (text,
				text->hadj,
				GTK_VALUE_POINTER (*arg));
      break;
	default:
		break;
    }
}

static void
gtk_extext_get_arg (GtkObject        *object,
		  GtkArg           *arg,
		  guint             arg_id)
{
  GtkExText *text;
  
  text = GTK_EXTEXT (object);
  
  switch (arg_id)
    {
    case ARG_HADJUSTMENT:
      GTK_VALUE_POINTER (*arg) = text->hadj;
      break;
    case ARG_VADJUSTMENT:
      GTK_VALUE_POINTER (*arg) = text->vadj;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
    }
}

static void gtk_extext_class_init(GtkExTextClass *class)
{
 GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkEditableClass *editable_class;
  
  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  editable_class = (GtkEditableClass*) class;
  parent_class = gtk_type_class (GTK_TYPE_EDITABLE);

   extext_signals[PROPERTY_INSERT]=gtk_signal_new ("property_insert",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkExTextClass, property_insert),
		    gtk_marshal_BOOL__POINTER_INT_INT,
		    GTK_TYPE_BOOL, 3,GTK_TYPE_POINTER,GTK_TYPE_INT,GTK_TYPE_INT);

   extext_signals[PROPERTY_REMOVE]=gtk_signal_new ("property_remove",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkExTextClass, property_remove),
		    gtk_marshal_BOOL__POINTER_INT_INT,
		    GTK_TYPE_BOOL, 3,GTK_TYPE_POINTER,GTK_TYPE_INT,GTK_TYPE_INT);

   extext_signals[UNDO_CHANGED]=gtk_signal_new ("undo_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkExTextClass, undo_changed),
		    gtk_marshal_NONE__NONE,
		    GTK_TYPE_NONE, 0);

	/* called if cursor has moved to a new property */
   extext_signals[PROPERTY_MARK]=gtk_signal_new ("property_mark",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkExTextClass, property_mark),
		    gtk_marshal_NONE__NONE,
		    GTK_TYPE_NONE, 0);

   extext_signals[LINE_INSERT]=gtk_signal_new ("line_insert",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkExTextClass, line_insert),
		    gtk_marshal_NONE__INT,
		    GTK_TYPE_NONE, 1, GTK_TYPE_INT);

   extext_signals[LINE_REMOVE]=gtk_signal_new ("line_remove",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkExTextClass, line_remove),
		    gtk_marshal_NONE__INT,
		    GTK_TYPE_NONE, 1,GTK_TYPE_INT);
	
  gtk_object_class_add_signals (object_class, extext_signals, LAST_SIGNAL);

  gtk_object_add_arg_type ("GtkExText::hadjustment",
			   GTK_TYPE_ADJUSTMENT,
			   GTK_ARG_READWRITE | GTK_ARG_CONSTRUCT,
			   ARG_HADJUSTMENT);
  gtk_object_add_arg_type ("GtkExText::vadjustment",
			   GTK_TYPE_ADJUSTMENT,
			   GTK_ARG_READWRITE | GTK_ARG_CONSTRUCT,
			   ARG_VADJUSTMENT);

  object_class->set_arg = gtk_extext_set_arg; 
  object_class->get_arg = gtk_extext_get_arg; 

  object_class->destroy=gtk_extext_destroy;
  object_class->finalize=gtk_extext_finalize;



  widget_class->size_request=gtk_extext_size_request;
  widget_class->size_allocate=gtk_extext_size_allocate;
  widget_class->realize=gtk_extext_realize;
  widget_class->unrealize=gtk_extext_unrealize;
  widget_class->button_press_event = gtk_extext_button_press;
  widget_class->button_release_event = gtk_extext_button_release;
	widget_class->motion_notify_event=gtk_extext_motion_notify;
  widget_class->draw = gtk_extext_draw;
  widget_class->draw_focus=gtk_extext_draw_focus;
  widget_class->expose_event = gtk_extext_expose;
  widget_class->key_press_event=gtk_extext_key_press;
  widget_class->focus_in_event = gtk_extext_focus_in;
  widget_class->focus_out_event = gtk_extext_focus_out;
	

	editable_class->insert_text=gtk_extext_insert_text;
	editable_class->delete_text=gtk_extext_delete_text;
/*	editable_class->changed=gtk_extext_changed;*/
	editable_class->set_position=gtk_extext_set_position;	
	editable_class->update_text=gtk_extext_update_text;
/*	editable_class->get_position=gtk_extext_get_position;	*/
	editable_class->get_chars=gtk_extext_get_chars;
	editable_class->set_editable=gtk_extext_set_editable;
	editable_class->set_selection=gtk_extext_select_region;
	
	/*editable_class->visible=TRUE;
*/

	class->property_insert=NULL;
	class->property_remove=NULL;
     class->undo_changed=NULL;
	class->property_mark=NULL;
	class->line_insert=NULL;
	class->line_remove=NULL;
}
static void
clear_focus_area (GtkExText *text, gint area_x, gint area_y, gint area_width, gint area_height)
{
  GtkWidget *widget = GTK_WIDGET (text);
  
/*  gint ythick = TEXT_BORDER_ROOM + widget->style->klass->ythickness;*/
/*  gint xthick = TEXT_BORDER_ROOM + widget->style->klass->xthickness;*/
  
  gint width, height;
  
  gdk_window_get_size (widget->style->bg_pixmap[GTK_STATE_NORMAL], &width, &height);
  
  /*gdk_gc_set_ts_origin (text->bg_gc,
			(- text->first_onscreen_hor_pixel + xthick) % width,
			(- text->first_onscreen_ver_pixel + ythick) % height);
*/
  gdk_draw_rectangle (GTK_WIDGET (text)->window, text->bg_gc, TRUE,
		      area_x, area_y, area_width, area_height);
}

void gtk_extext_draw_focus (GtkWidget *widget)
{
  GtkExText *text;
  gint width, height;
  gint x, y;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (widget));
  
  text = GTK_EXTEXT (widget);
  
  if (GTK_WIDGET_DRAWABLE (widget))
    {
      gint ythick = widget->style->klass->ythickness;
      gint xthick = widget->style->klass->xthickness;
      gint xextra = TEXT_BORDER_ROOM;
      gint yextra = TEXT_BORDER_ROOM;
      
      
      x = 0;
      y = 0;
      width = widget->allocation.width;
      height = widget->allocation.height;
      
      if (GTK_WIDGET_HAS_FOCUS (widget))
	{
	  x += 1;
	  y += 1;
	  width -=  2;
	  height -= 2;
	  xextra -= 1;
	  yextra -= 1;

	  gtk_paint_focus (widget->style, widget->window,
			   NULL, widget, "text",
			   0, 0,
			   widget->allocation.width - 1,
			   widget->allocation.height - 1);
	}

      gtk_paint_shadow (widget->style, widget->window,
			GTK_STATE_NORMAL, GTK_SHADOW_IN,
			NULL, widget, "text",
			x, y, width, height);

      x += xthick; 
      y += ythick;
      width -= 2 * xthick;
      height -= 2 * ythick;
      
      if (widget->style->bg_pixmap[GTK_STATE_NORMAL])
	{
	  /* top rect */
	  clear_focus_area (text, x, y, width, yextra);
	  /* left rect */
	  clear_focus_area (text, x, y + yextra, 
			    xextra, y + height - 2 * yextra);
	  /* right rect */
	  clear_focus_area (text, x + width - xextra, y + yextra, 
			    xextra, height - 2 * ythick);
	  /* bottom rect */
	  clear_focus_area (text, x, x + height - yextra, width, yextra);
	}
    }
  else
    {
      g_print (("in gtk_text_draw_focus (undrawable !!!)\n"));
    }
}

static void
gtk_extext_size_request (GtkWidget      *widget,
		       GtkRequisition *requisition)
{
  gint xthickness;
  gint ythickness;
  gint char_height;
  gint char_width; 
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (widget));
  g_return_if_fail (requisition != NULL);
     xthickness = widget->style->klass->xthickness + TEXT_BORDER_ROOM;
  ythickness = widget->style->klass->ythickness + TEXT_BORDER_ROOM;
  
  char_height = MIN_TEXT_HEIGHT_LINES * (widget->style->font->ascent +
					 widget->style->font->descent);
  
  char_width = MIN_TEXT_WIDTH_LINES * (gdk_text_width (widget->style->font,
						       "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
						       26) 
				       / 26);
  
  requisition->width  = char_width  + xthickness * 2;
  requisition->height = char_height + ythickness * 2;
}
static void
gtk_extext_size_allocate (GtkWidget     *widget,
			GtkAllocation *allocation)
{
  GtkExText *text;
  GtkEditable *editable;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (widget));
  g_return_if_fail (allocation != NULL);
  
  text = GTK_EXTEXT (widget);
  editable = GTK_EDITABLE (widget);
  
  widget->allocation = *allocation;
  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);
      
      gdk_window_move_resize (text->text_area,
			      widget->style->klass->xthickness + TEXT_BORDER_ROOM,
			      widget->style->klass->ythickness + TEXT_BORDER_ROOM,
			      MAX (1, (gint)widget->allocation.width - (gint)(widget->style->klass->xthickness +
							  (gint)TEXT_BORDER_ROOM) * 2),
			      MAX (1, (gint)widget->allocation.height - (gint)(widget->style->klass->ythickness +
							   (gint)TEXT_BORDER_ROOM) * 2));
      
#ifdef USE_XfIM
      if (editable->ic && (gdk_ic_get_style (editable->ic) & GDK_IM_PREEDIT_POSITION))
	{
	  gint width, height;
	  
	  gdk_window_get_size (text->text_area, &width, &height);
	  editable->ic_attr->preedit_area.width = width;
	  editable->ic_attr->preedit_area.height = height;

	  gdk_ic_set_attr (editable->ic,
	      		   editable->ic_attr, GDK_IC_PREEDIT_AREA);
	}
#endif
      
      recompute_geometry (text);
    }
}
	/* vertical and horizontal scroll should be updated here */
	
static void recompute_geometry (GtkExText* text)
{	
	gint height,width;
	GtkExTextStyle *style;

	  gdk_window_get_size (text->text_area, &width, &height);

	style=gtk_extext_style_get(text,"Default");
	text->scroll_line_count=height/get_real_text_height(style->font);

	if(text->scroll_line_count) {
		text->vadj->page_size=text->line_count/text->scroll_line_count;
		text->scroll_line_count--;
	}
	else
		text->vadj->page_size=0;


	text->vadj->step_increment=1.0;
	text->vadj->page_increment=text->scroll_line_count;
	text->vadj->page_size=text->scroll_line_count;
	text->vadj->upper=text->line_count;


	gtk_adjustment_changed(text->vadj);

/*
	g_print("update scroll_line_count %d height %d\n",text->scroll_line_count,height); */
}

void
gtk_extext_set_adjustments (GtkExText       *text,
			  GtkAdjustment *hadj,
			  GtkAdjustment *vadj)
{
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));
  if (hadj)
    g_return_if_fail (GTK_IS_ADJUSTMENT (hadj));
  else
    hadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
  if (vadj)
    g_return_if_fail (GTK_IS_ADJUSTMENT (vadj));
  else
    vadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
  
  if (text->hadj && (text->hadj != hadj))
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (text->hadj), text);
      gtk_object_unref (GTK_OBJECT (text->hadj));
    }
  
  if (text->vadj && (text->vadj != vadj))
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (text->vadj), text);
      gtk_object_unref (GTK_OBJECT (text->vadj));
    }
  
  if (text->hadj != hadj)
    {
      text->hadj = hadj;
      gtk_object_ref (GTK_OBJECT (text->hadj));
      gtk_object_sink (GTK_OBJECT (text->hadj));
      
      gtk_signal_connect (GTK_OBJECT (text->hadj), "changed",
			  (GtkSignalFunc) gtk_extext_adjustment_changed,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->hadj), "value_changed",
			  (GtkSignalFunc) gtk_extext_adjustment_value_changed,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->hadj), "disconnect",
			  (GtkSignalFunc) gtk_extext_disconnect,
			  text);
      gtk_extext_adjustment_changed (hadj, text);
    }
  
  if (text->vadj != vadj)
    {
      text->vadj = vadj;
      gtk_object_ref (GTK_OBJECT (text->vadj));
      gtk_object_sink (GTK_OBJECT (text->vadj));
      
      gtk_signal_connect (GTK_OBJECT (text->vadj), "changed",
			  (GtkSignalFunc) gtk_extext_adjustment_changed,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->vadj), "value_changed",
			  (GtkSignalFunc) gtk_extext_adjustment_value_changed,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->vadj), "disconnect",
			  (GtkSignalFunc) gtk_extext_disconnect,
			  text);
      gtk_extext_adjustment_changed (vadj, text);
    }
	
	text->hadj->step_increment=1.0;
	text->hadj->page_increment=50.0;
	text->hadj->page_size=50.0;
	text->hadj->upper=200.0;
	text->hadj->value=0.0;
	
	
	text->vadj->step_increment=1.0;
	text->vadj->page_increment=text->scroll_line_count;
	text->vadj->page_size=text->scroll_line_count;
	text->vadj->upper=text->line_count;
}


static void gtk_extext_init(GtkExText *text)
{
	GTK_WIDGET_SET_FLAGS(text,GTK_CAN_FOCUS);
	text->gc=NULL;
	text->bg_gc=NULL;
	text->text_area=NULL;
	text->font=NULL;


	text->text=g_new0(guchar,TEXT_BUFFERT_DEFAULT_SIZE);
	text->size=TEXT_BUFFERT_DEFAULT_SIZE;	
	text->gap_len=TEXT_BUFFERT_DEFAULT_SIZE;
	text->part1len=0;
	text->part2text=text->text+text->gap_len;
	text->length=0;

	text->read_only=FALSE;

	text->line_start=text->line_end=g_malloc0(sizeof(LineCache));
	text->line_pos=text->line_start;
	text->line_pos_index=0;
	text->line_count=0;
	text->line_cursor_index=0;

	text->hash_styles=g_hash_table_new(g_str_hash,g_str_equal);
	text->property_start=NULL;
	text->property_end=NULL;
	text->property_current=NULL;

	text->undo_buffert=NULL;
	text->undo_max=-1;
	text->undo_count=0;
	text->undo_last=0;
	text->undo_flag=FALSE;

	GTK_EDITABLE(text)->current_pos=0;

}

GtkWidget * gtk_extext_new(GtkAdjustment *hadj,GtkAdjustment *vadj)
{
	GtkWidget *text;

	if (hadj)
 	   g_return_val_if_fail (GTK_IS_ADJUSTMENT (hadj), NULL);
 	if (vadj)
  	  g_return_val_if_fail (GTK_IS_ADJUSTMENT (vadj), NULL);

	  text = gtk_widget_new (GTK_TYPE_EXTEXT,
			 "hadjustment", hadj,
			 "vadjustment", vadj,
			 NULL);


	return text;
}

static void gtk_extext_destroy(GtkObject *object)
{
	GtkExText *text;
	LineCache *current,*next;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (object));
  
	text=GTK_EXTEXT(object);
  
	gtk_extext_undo_clear_all(GTK_EXTEXT(object));
	property_class_destroy(text);

	g_free(text->text);
	current=text->line_start;
	while(current)
	{
		next=current->next;
		g_free(current);
		current=next;
	};	


  gtk_signal_disconnect_by_data (GTK_OBJECT (text->hadj), text);
  gtk_signal_disconnect_by_data (GTK_OBJECT (text->vadj), text);

  if (text->timer)
    {
      gtk_timeout_remove (text->timer);
      text->timer = 0;
    }
  
	GTK_OBJECT_CLASS(parent_class)->destroy(object);
}
static void gtk_extext_finalize(GtkObject *text)
{

}
static void gtk_extext_realize(GtkWidget *widget)
{
  GtkExText *text;
  GtkEditable *editable;
  GdkWindowAttr attributes;
  gint attributes_mask;
	GtkExTextStyle *style;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (widget));
  
  text = GTK_EXTEXT (widget);
  editable = GTK_EDITABLE (widget);
  GTK_WIDGET_SET_FLAGS (text, GTK_REALIZED);
  
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_BUTTON_PRESS_MASK |
			    GDK_BUTTON_RELEASE_MASK |
			    GDK_BUTTON_MOTION_MASK |
			    GDK_ENTER_NOTIFY_MASK |
			    GDK_LEAVE_NOTIFY_MASK |
			    GDK_KEY_PRESS_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  
  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, text);
  
  attributes.x = (widget->style->klass->xthickness + TEXT_BORDER_ROOM);
  attributes.y = (widget->style->klass->ythickness + TEXT_BORDER_ROOM);
  attributes.width = MAX (1, (gint)widget->allocation.width - (gint)attributes.x * 2);
  attributes.height = MAX (1, (gint)widget->allocation.height - (gint)attributes.y * 2);
  
  text->text_area = gdk_window_new (widget->window, &attributes, attributes_mask);
  gdk_window_set_user_data (text->text_area, text);
  
  widget->style = gtk_style_attach (widget->style, widget->window);
  
  /* Can't call gtk_style_set_background here because it's handled specially */
  gdk_window_set_background (widget->window, &widget->style->base[GTK_STATE_NORMAL]);
  gdk_window_set_background (text->text_area, &widget->style->base[GTK_STATE_NORMAL]);

/*  if (widget->style->bg_pixmap[GTK_STATE_NORMAL])
    text->bg_gc = create_bg_gc (text);*/
  
/*  text->cache_wrap_bitmap = gdk_bitmap_create_from_data (text->text_area,
							(gchar*) cache_wrap_bits,
							cache_wrap_width,
							cache_wrap_height);
  
  text->cache_arrow_bitmap = gdk_bitmap_create_from_data (text->text_area,
							 (gchar*) cache_arrow_bits,
							 cache_arrow_width,
							 cache_arrow_height);
*/  

  text->bg_gc = gdk_gc_new (text->text_area);
  gdk_gc_set_foreground (text->bg_gc, &widget->style->base[GTK_STATE_SELECTED]);
  gdk_gc_set_background (text->bg_gc, &widget->style->base[GTK_STATE_SELECTED]);

  text->gc = gdk_gc_new (text->text_area);
  gdk_gc_set_exposures (text->gc, TRUE);
  gdk_gc_set_foreground (text->gc, &widget->style->text[GTK_STATE_NORMAL]);
  
#ifdef USE_XIM
  if (gdk_im_ready () && (editable->ic_attr = gdk_ic_attr_new ()) != NULL)
    {
      gint width, height;
      GdkColormap *colormap;
      GdkEventMask mask;
      GdkICAttr *attr = editable->ic_attr;
      GdkICAttributesType attrmask = GDK_IC_ALL_REQ;
      GdkIMStyle style;
      GdkIMStyle supported_style = GDK_IM_PREEDIT_NONE | 
	                           GDK_IM_PREEDIT_NOTHING |
	                           GDK_IM_PREEDIT_POSITION |
	                           GDK_IM_STATUS_NONE |
	                           GDK_IM_STATUS_NOTHING;
      
      if (widget->style && widget->style->font->type != GDK_FONT_FONTSET)
	supported_style &= ~GDK_IM_PREEDIT_POSITION;
      
      attr->style = style = gdk_im_decide_style (supported_style);
      attr->client_window = text->text_area;

      if ((colormap = gtk_widget_get_colormap (widget)) !=
	  gtk_widget_get_default_colormap ())
	{
	  attrmask |= GDK_IC_PREEDIT_COLORMAP;
	  attr->preedit_colormap = colormap;
	}

      switch (style & GDK_IM_PREEDIT_MASK)
	{
	case GDK_IM_PREEDIT_POSITION:
	  if (widget->style && widget->style->font->type != GDK_FONT_FONTSET)
	    {
	      g_warning ("over-the-spot style requires fontset");
	      break;
	    }

	  attrmask |= GDK_IC_PREEDIT_POSITION_REQ;
	  gdk_window_get_size (text->text_area, &width, &height);
	  attr->spot_location.x = 0;
	  attr->spot_location.y = height;
	  attr->preedit_area.x = 0;
	  attr->preedit_area.y = 0;
	  attr->preedit_area.width = width;
	  attr->preedit_area.height = height;
	  attr->preedit_fontset = widget->style->font;
	  
	  break;
	}
      editable->ic = gdk_ic_new (attr, attrmask);
      
      if (editable->ic == NULL)
	g_warning ("Can't create input context.");
      else
	{
	  mask = gdk_window_get_events (text->text_area);
	  mask |= gdk_ic_get_events (editable->ic);
	  gdk_window_set_events (text->text_area, mask);
	  
	  if (GTK_WIDGET_HAS_FOCUS (widget))
	    gdk_im_begin (editable->ic, text->text_area);
	}
    }
#endif

	text->font=widget->style->font;
  gdk_window_show (text->text_area);

  if (editable->selection_start_pos != editable->selection_end_pos)
    gtk_editable_claim_selection (editable, TRUE, GDK_CURRENT_TIME);


	
	style=gtk_extext_style_insert(text,"Default",
					GTK_WIDGET(text)->style->font,
					&GTK_WIDGET(text)->style->fg[GTK_STATE_NORMAL],
					&GTK_WIDGET(text)->style->white	);


	gtk_extext_style_insert(text,"Selected",
					GTK_WIDGET(text)->style->font,
		&GTK_WIDGET(text)->style->fg[GTK_STATE_SELECTED],
					&GTK_WIDGET(text)->style->bg[GTK_STATE_SELECTED]	);

	gtk_extext_style_insert(text,"Prelight",
					GTK_WIDGET(text)->style->font,
		&GTK_WIDGET(text)->style->fg[GTK_STATE_PRELIGHT],
					&GTK_WIDGET(text)->style->bg[GTK_STATE_PRELIGHT]	);

	text->timer=0;
  text->freeze=FALSE;

  recompute_geometry (text);
}
static void gtk_extext_unrealize(GtkWidget *widget)
{
 GtkExText *text;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (widget));
  
  text = GTK_EXTEXT (widget);

  gdk_window_set_user_data (text->text_area, NULL);
  gdk_window_destroy (text->text_area);
  text->text_area = NULL;
  
  gdk_gc_destroy (text->gc);
  text->gc = NULL;

  if (text->bg_gc)
    {
      gdk_gc_destroy (text->bg_gc);
      text->bg_gc = NULL;
    }
  
  if (GTK_WIDGET_CLASS (parent_class)->unrealize)
    (* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
}
static void expose_text(GtkExText *text,GdkRectangle *area,gboolean hmm)
{
	gint lines;
	gint end;	
	gint i;
	gint w,h;
	GdkDrawable *draw_area;
	gint line_start_pos;
	LineCache *cur;
	GtkExTextProperty *prop;	

	if(text->freeze)
		return;

	lines=text->scroll_line_start;
	if(text->scroll_line_count+text->scroll_line_start>text->line_count)
		end=text->line_count;
	else
		end=text->scroll_line_start+text->scroll_line_count;

	i=0;

	/* THIS IS A VERY QUICK HACK THIS WILL FLICKER */
	/*ALOT everytime GTK REPAINTS */
	
/* Fix later :-/ */
	
	gdk_window_get_size(text->text_area,&w,&h);

	w=gdk_char_width(text->font,' ');

	draw_area=gdk_pixmap_new(text->text_area,w*MAX_LINE_LEN,h,-1);
	gdk_draw_rectangle(draw_area,
				GTK_WIDGET(text)->style->white_gc
					,TRUE,0,0,w*MAX_LINE_LEN,h);


	lines=text->scroll_line_start;
	cur=line_get_ptr(text,&lines);
	line_start_pos=line_get_start_pos(text,cur);
	prop=property_get_nearest_at_pos(text,line_start_pos,FALSE,NULL);
	for(;lines<=end;lines++)	{
		expose_parse_line_text(text,draw_area,lines,cur,line_start_pos,prop,0,0);
		line_start_pos+=cur->len;
		cur=cur->next;
	}


	if(draw_area!=text->text_area)
	{
		gdk_draw_pixmap(text->text_area,
					GTK_WIDGET(text)->style->fg_gc[GTK_WIDGET_STATE(text)],
						draw_area,w*text->hadj->value+area->x,area->y,area->x,area->y,
							area->width,area->height	);

		gdk_pixmap_unref(draw_area);
	}
	gtk_widget_draw_focus(GTK_WIDGET(text)); 
}
void expose_parse_line_text(GtkExText *text,GdkDrawable *draw_area,
					gint line,LineCache *lptr,gint startpos,
					GtkExTextProperty *oldprop,gint x,gint y)
{
		gchar *txt;
		gint cursortabcount;
		gint i,pindex,endpos;
		gint tabindex,tab;	
		GtkExTextStyle *style; 
		GtkExTextStyle *style_selected; 
		gint height;
		gint nheight;
		gint offset;
		GtkExTextProperty *prop; 
		
		/* hopefully there is not more than 64 propertys in one line  :-( */
		/* if there is the text will not appear */
		/* This is a very dirty hack */
		/* maybe we should do it better in the feature */

		InternalPaintCache pcache[64];

		prop=NULL;
		offset=0;
		tabindex=0;					/* TODO a tabtable */
		tab=0;				/* is there a tab this is set to [tabindex].tab */
		pindex=0;				/* pcache index */
		style=0;						/* Current style */
		cursortabcount=0;				/* tabs before cursor pos */ 

		endpos=startpos+lptr->len;
		
		txt=text_get_text(text,startpos,endpos);
		if(!txt)
			return;

		/* Cache 0 is ALWAYS Default */
  
		style=pcache[0].style=gtk_extext_style_get(text,"Default");		
		pcache[0].len=0;

		
		height=get_real_text_height(pcache[0].style->font);

		if(GTK_WIDGET_HAS_FOCUS(text))
			style_selected=gtk_extext_style_get(text,"Selected");
		else
			style_selected=gtk_extext_style_get(text,"Prelight");

		for( i=startpos;i<endpos;i++)
		{
			if(pindex>59)
					break;

			tab=0;	
			if( txt[i-startpos]=='\t')	{	
				tab=5 -1;		/*tabspace=tab_table[tabindex] */
											/*-tabtable[tabindex-1]; FIXLATER */
				
				if(text->line_pos==lptr 
							&& i - startpos<text->line_cursor_index 
									&& text->line_cursor_index)
					cursortabcount+=tab;
				
				tabindex++;
			}

			
			
			prop=gtk_extext_property_get_at_pos(text,i,oldprop);
			if(prop)	{
				style=gtk_extext_style_get(text,prop->style_key);
				oldprop=prop;
			}
			else
					style=pcache[0].style; 

			/* SELECTIONS ALWAYS CLEAR STYLE AT POS */
			if(!is_value_ok(i, GTK_EDITABLE(text)->selection_start_pos,GTK_EDITABLE(text)->selection_end_pos))	{
				if(style_selected!=pcache[pindex].style)	{
					pindex++;
					pcache[pindex].style=style_selected;

						pcache[pindex].len=tab+1;

					nheight=get_real_text_height(style->font);
					if(nheight>height)
						height=nheight;
				}
				else
				{	
					pcache[pindex].len+=tab+1;
				}
			}	
			/* Init new style */
			else if(style!=pcache[pindex].style){
				pindex++;
				pcache[pindex].style=style;

				pcache[pindex].len=tab+1;
				nheight=get_real_text_height(style->font);
				if(nheight>height)
					height=nheight;
			}
			else	{
				pcache[pindex].len+=tab+1;				
			}
		}	
		
	
	g_free(txt);
	
	tabindex=0;
	txt=text_get_text_with_expanded_tabs(text,startpos,endpos,&tabindex);

	if(txt)	{
		x=0;
		y=line - text->scroll_line_start;
		offset=0;
		for(i=0;i<=pindex;i++)	{
			if(pcache[i].len)
			{
				gint width;
				GtkExTextStyle *style;
				gint descent;

				style=pcache[i].style;

				descent=get_text_descent(style->font);
				width=gdk_text_width(style->font,
								&txt[offset],pcache[i].len);

				gdk_gc_set_foreground(text->gc,&style->bg_color);
				gdk_draw_rectangle(draw_area,text->gc,TRUE,
							x,height*y,width,height);      

				gdk_gc_set_foreground(text->gc,&style->fg_color);
				gdk_draw_text(draw_area,style->font,text->gc,
								x,(y+1)* height - descent,&txt[offset],pcache[i].len);
			
				x+=width;
				offset+=pcache[i].len;

			}
		}
		if(lptr==text->line_pos)
		{
			gint width;
	
			width=gdk_text_width(pcache[0].style->font,txt,
									text->line_cursor_index+cursortabcount);
			
			gdk_draw_line(draw_area,GTK_WIDGET(text)->style->black_gc,
					width,y*height,width,y * height + height);
		}

		g_free(txt);
	}
}
static void
gtk_extext_disconnect (GtkAdjustment *adjustment,
		     GtkExText       *text)
{
  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));

  if (adjustment == text->hadj)
    gtk_extext_set_adjustments (text, NULL, text->vadj);
  if (adjustment == text->vadj)
    gtk_extext_set_adjustments (text, text->hadj, NULL);
}

static gint
gtk_extext_scroll_timeout (gpointer data)
{
  GtkExText *text;
  GdkEventMotion event;
  gint x, y;
  GdkModifierType mask;
  
  GDK_THREADS_ENTER ();

  text = GTK_EXTEXT (data);
  
  text->timer = 0;
  gdk_window_get_pointer (text->text_area, &x, &y, &mask);
  
  if (mask & (GDK_BUTTON1_MASK | GDK_BUTTON3_MASK))
    {
      event.is_hint = 0;
      event.x = x;
      event.y = y;
      event.state = mask;
      
      gtk_extext_motion_notify (GTK_WIDGET (text), &event);
    }

  GDK_THREADS_LEAVE ();
  
  return FALSE;
}

static void gtk_extext_delete_to_line_end(GtkExText *text)
{
	GtkEditable *editable;
	gint len;
  g_return_if_fail (GTK_IS_EXTEXT (text));
	
	editable=GTK_EDITABLE(text);

	len=editable->current_pos - line_get_start_pos(text,text->line_pos);
 	gtk_editable_delete_text(editable,editable->current_pos,
								editable->current_pos + text->line_pos->len - len);	
}
static void gtk_extext_delete_backward_word(GtkExText *text)
{
	gint start,end;
	GtkEditable *editable;
  g_return_if_fail (GTK_IS_EXTEXT (text));
	editable=GTK_EDITABLE(text);

	start=end=editable->current_pos;
	gtk_extext_get_previous_word(text,&start,&end);		
	gtk_editable_select_region(editable,start,editable->current_pos);
	gtk_editable_delete_selection(editable); /* ,start,editable->current_pos); */
}
static void gtk_extext_delete_line(GtkExText *text)
{
	gint start;
	GtkEditable *editable;
  g_return_if_fail (GTK_IS_EXTEXT (text));

	editable=GTK_EDITABLE(text);
	start=line_get_start_pos(text,text->line_pos);
	gtk_editable_delete_text(editable,start,
								start + text->line_pos->len);	

}



/*****************************************************************************/
/*                    Mousecursor TO line and linecursorpos                      */
/*                         Code 																*/
/*****************************************************************************/

static void find_cursor(GtkExText *text,gdouble x,gdouble y,gboolean select)
{
	GtkExTextStyle *style;
	gint height,width;
	gint start,end;
	LineCache* lineptr;
	gint cursor;
	guchar *txttab; 
	guchar key;
	gint tabs;
	gint tabindex;
	gint linestart;
	gint i;
				
	i=0;
	tabindex=0;
	tabs=0;		
	style=gtk_extext_style_get(text,"Default");
	height=get_real_text_height(style->font);

	/* Oki lets see were the fucking cursor is */

	lineptr=NULL;
	cursor=-1;


	if(y<0){	
			lineptr=goto_line(text,text->scroll_line_start-1,select,TRUE);
		  text->timer = gtk_timeout_add (SCROLL_TIME, 
					 gtk_extext_scroll_timeout,
					 text);
	}
	else	{
		for(i=0;i<text->scroll_line_count;i++)	{
			if(y <=((float)height*(i+1)) )	{
				lineptr=goto_line(text,i+text->scroll_line_start,select,TRUE);
				break;
			}
		}
	/* Oki we no witch line but now its little more tricky to now were the cursor is*/

	}
	if(!lineptr)  { /* Happens if out of scrollview */
		lineptr=goto_line(text,i+text->scroll_line_start,select,TRUE);
		  text->timer = gtk_timeout_add (SCROLL_TIME, 
					 gtk_extext_scroll_timeout,
					 text);
	}


	linestart=line_get_start_pos(text,lineptr);
	start=linestart + text->hadj->value;
	end=linestart + lineptr->len;
	txttab=NULL;

	/* happens if line is less than hscroll */
	if(end<start)
		move_cursor(text,lineptr->len - 1  ,select);	
	else {
		txttab=text_get_text_with_expanded_tabs(text,start,end,&tabindex);
		for(i=0;txttab && i<strlen(txttab);i++)	{
			if(i<end)	{
				key=gtk_extext_get_char_at_pos(text,i+start);
				if(key=='\t')		
					tabs+=4;
			}
			width=gdk_text_width(style->font,txttab,i+tabs+1);		
			if(x <= (float)width )	{
				cursor=i + text->hadj->value;
				break;
			}
		}
		if(txttab)			
			g_free(txttab);

		if(cursor==-1)
			move_cursor(text, lineptr->len-1 ,select);
		else
			move_cursor(text,cursor,select);
	}

	gtk_widget_queue_draw(GTK_WIDGET(text));
}
static void gtk_extext_draw(GtkWidget *widget,GdkRectangle *area)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (widget));
  g_return_if_fail (area != NULL);
  
  if (GTK_WIDGET_DRAWABLE (widget))
    {
      expose_text (GTK_EXTEXT (widget), area, TRUE);
      gtk_widget_draw_focus (widget);
    }

}
/*****************************************************************************/
/*                              extext_key_press                                */
/*****************************************************************************/

static gint  gtk_extext_key_press         (GtkWidget         *widget,
					 GdkEventKey       *event)
{
	GtkExText *text;
	GtkEditable *editable;
	gboolean shift_state;
	gboolean paint;
	gboolean return_value;
	gchar key;
	gint curpos;  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
	text=GTK_EXTEXT(widget);
	editable=GTK_EDITABLE(widget);
	paint=FALSE;


	curpos=editable->current_pos;

	return_value=FALSE;

	shift_state=FALSE;
	if (event->state & GDK_SHIFT_MASK)
		shift_state=TRUE;

	
	key=event->keyval;
	  if (event->state & GDK_CONTROL_MASK)
	 {
	      if ((key >= 'A') && (key <= 'Z'))
			key -= 'A' - 'a';
	      
	      if ((key >= 'a') && (key <= 'z') && control_keys[(int) (key - 'a')])
		{
		  (* control_keys[(int) (key - 'a')]) (editable, event->time);
		  return_value = TRUE;
			paint=TRUE;
		}
		switch(event->keyval) 
		{		
			case GDK_Home:	
				gtk_extext_set_line(text,0);
				return_value=TRUE;
			break;
			case GDK_End:
				gtk_extext_set_line(text,text->line_count);
				return_value=TRUE;
			break;
		}
      
	 }
	else	if(!(event->state & GDK_MOD1_MASK)){
		return_value=TRUE;
		paint=TRUE;
		switch(event->keyval)
		{
			case GDK_Escape:
				paint=TRUE;
			break;
			case GDK_Home:
			    	line_set_cursor_pos(text,0);				
			break;
			case GDK_End:
				if(text->line_pos==text->line_end)
					line_set_cursor_pos(text,text->line_pos->len);	
				else
					line_set_cursor_pos(text,text->line_pos->len-1);				
			break;
			case GDK_Page_Up:
				goto_line(text,line_get_num(text,text->line_pos,FALSE) 
										- text->scroll_line_count,shift_state,TRUE);
			break;
			case GDK_Page_Down:
				goto_line(text,line_get_num(text,text->line_pos,TRUE)
										+ text->scroll_line_count,shift_state,TRUE);
			break;
			case GDK_Up:
				goto_line(text,text->line_pos_index-1,shift_state,TRUE);
			break;
			case GDK_Down:
				goto_line(text,text->line_pos_index+1,shift_state,TRUE);
			break;
			case GDK_Left:
				move_cursor(text, text->line_cursor_index  -1,shift_state );
				paint=TRUE;
			break;
			case GDK_Right:
				move_cursor(text, text->line_cursor_index +1 ,shift_state);
				paint=TRUE;
			break;
			case GDK_Return:
				gtk_editable_insert_text(editable,"\n",1,&curpos);
			break;
			case  GDK_KP_Enter:
				gtk_editable_insert_text(editable,"\n",1,&curpos);
			break;
			case GDK_BackSpace:
				gtk_editable_delete_text(GTK_EDITABLE(text),
							GTK_EDITABLE(text)->current_pos-1,
							GTK_EDITABLE(text)->current_pos);
			break;
			case GDK_Delete:
				gtk_editable_delete_text(GTK_EDITABLE(text),
							GTK_EDITABLE(text)->current_pos,
							GTK_EDITABLE(text)->current_pos+1);
			break;
			case GDK_Tab:
				gtk_editable_insert_text(editable,"\t",1,&curpos);
			break;
			default:
				if(event->length)
					gtk_editable_insert_text(editable,(guchar*)event->string,event->length,&curpos);
				else 
					return_value=FALSE; /* let the parentwidget take care */
			break;
		};
	}

	if(paint)
		gtk_widget_queue_draw(widget);

	return return_value;
}

static void
gtk_extext_adjustment_changed (GtkAdjustment *adjustment,
		     GtkExText       *text)
{
  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));
  
  /* Just ignore it if we haven't been size-allocated and realized yet */
  
  if (adjustment == text->hadj)
    {
    }
  else if(adjustment==text->vadj)
    {
		text->vadj->upper=(float)text->line_count;	
		text->scroll_line_start=(gint)text->vadj->value;
	/*	gtk_widget_queue_draw(GTK_WIDGET(text));
*/
    }
}
static void gtk_extext_adjustment_value_changed (GtkAdjustment *adjustment,
		     GtkExText       *text)
{
	gint lineto;
	gint pos;

/*
	g_print("extext_adjustemnt_value_changed\n");
*/
	lineto=(gint)text->vadj->value;

	if(text->vadj->value>=text->line_pos_index)
		pos=line_get_num(text,text->line_pos,TRUE);
	else
		pos=line_get_num(text,text->line_pos,FALSE);

	if(pos<lineto)	{
		goto_line(text,lineto,FALSE,FALSE);
	}
	/* if out o
f viewport */
	if(pos >= lineto + text->scroll_line_count )	{
						goto_line(text,
									lineto + text->scroll_line_count-1,FALSE,FALSE);
	}
	text->scroll_line_start=lineto;



	gtk_widget_queue_draw(GTK_WIDGET(text));
}

static gint
gtk_extext_focus_in (GtkWidget     *widget,
		   GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  
  
  GTK_WIDGET_SET_FLAGS (widget, GTK_HAS_FOCUS);
  gtk_widget_draw_focus (widget);
      
  return FALSE;
}

static gint
gtk_extext_focus_out (GtkWidget     *widget,
		    GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  
  
  GTK_WIDGET_UNSET_FLAGS (widget, GTK_HAS_FOCUS);
  gtk_widget_draw_focus (widget);
	gtk_widget_queue_draw(widget);
    
  return FALSE;
}
static gint
gtk_extext_motion_notify (GtkWidget      *widget,
			GdkEventMotion *event)
{
  GtkExText *text;
  gint x, y;
  gint height;
  GdkModifierType mask;
  
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  
  text = GTK_EXTEXT (widget);
  
  x = event->x;
  y = event->y;
  mask = event->state;
  if (event->is_hint || (text->text_area != event->window))
    {
      gdk_window_get_pointer (text->text_area, &x, &y, &mask);
    }
  
  if ((text->button == 0) ||
      !(mask & (GDK_BUTTON1_MASK | GDK_BUTTON3_MASK)))
    return FALSE;
  
  gdk_window_get_size (text->text_area, NULL, &height);
  
 	find_cursor (GTK_EXTEXT (widget), (float)x,(float) y,TRUE);

  
  return FALSE;
}
static gint
gtk_extext_button_press (GtkWidget      *widget,
		       GdkEventButton *event)
{
	gint start,end;
  GtkExText *text;
  GtkEditable *editable;
	 static GdkAtom ctext_atom = GDK_NONE;


  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
 

 	if (ctext_atom == GDK_NONE)
		ctext_atom = gdk_atom_intern ("COMPOUND_TEXT", FALSE);
  

  text = GTK_EXTEXT (widget);
  editable = GTK_EDITABLE (widget);
  
  if (text->button && (event->button != text->button))
    return FALSE;
  
  text->button = event->button;
  
  if (!GTK_WIDGET_HAS_FOCUS (widget))
    gtk_widget_grab_focus (widget);
  
  if (event->button == 1)
    {
      switch (event->type)
	{
	case GDK_BUTTON_PRESS:
	  gtk_grab_add (widget);
	  
	  
	  /* Set it now, so we display things right. We'll unset it
	   * later if things don't work out */
	  editable->has_selection = TRUE;
	find_cursor(text,event->x,event->y,FALSE);
 
	  break;
	  
	case GDK_2BUTTON_PRESS:
		start=end=editable->current_pos;
		gtk_extext_get_previous_word(text,&start,&end);		
		gtk_editable_select_region(editable,start,end);
	break;
	  
	case GDK_3BUTTON_PRESS:
		start=line_get_start_pos(text,text->line_pos);
		gtk_editable_select_region(editable,start,start+text->line_pos->len);
	 break;
	  
	default:
	  break;
	}
    }
  else if (event->type == GDK_BUTTON_PRESS)
    {
      if ((event->button == 2) && editable->editable)
	{
	  if (editable->selection_start_pos == editable->selection_end_pos ||
	      editable->has_selection)
	    {
	      	g_print("do something\n");
	    }
	  
	  gtk_selection_convert (widget, GDK_SELECTION_PRIMARY,
				 ctext_atom, event->time);
	}
      else
	{
	  gtk_grab_add (widget);
	  
	  find_cursor (text, event->x, event->y,FALSE);
	  
	  
	  editable->has_selection = FALSE;
	  if (gdk_selection_owner_get (GDK_SELECTION_PRIMARY) == widget->window)
	    gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY, event->time);
	}
    }
  
	return FALSE;	
  
}

/*****************************************************************************/
/*                    button_release is stolen from GtkText                      */
/*****************************************************************************/


static gint
gtk_extext_button_release (GtkWidget      *widget,
			 GdkEventButton *event)
{
  GtkExText *text;
  GtkEditable *editable;
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  
  text = GTK_EXTEXT (widget);
  editable = GTK_EDITABLE (widget);

  gtk_grab_remove (widget);

  if (text->button != event->button)
    return FALSE;
 
  text->button = 0;
  
  if (text->timer)
    {
      gtk_timeout_remove (text->timer);
      text->timer = 0;
    }
  
   if (event->button == 1)
    {
      
      gtk_grab_remove (widget);
      
      editable->has_selection = FALSE;
      if (editable->selection_start_pos != editable->selection_end_pos)
	{
	  if (gtk_selection_owner_set (widget,
				       GDK_SELECTION_PRIMARY,
				       event->time))
	    editable->has_selection = TRUE;
	  else
		gtk_widget_queue_draw(widget);
	}
      else
	{
	  if (gdk_selection_owner_get (GDK_SELECTION_PRIMARY) == widget->window)
	    gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY, event->time);
	}
    }
  else if (event->button == 3)
    {
      gtk_grab_remove (widget);
    }

	return FALSE;
}
/* Make sure scrolladjustment is up to date */
void scroll_to_view(GtkExText *text,gint moveto)
{
	if(text->scroll_line_start+text->scroll_line_count<=moveto)	{
		text->scroll_line_start=moveto;
		text->vadj->value=(gfloat)moveto;
		gtk_adjustment_changed(text->vadj);
	}
	else if(text->scroll_line_start>moveto)	{
		text->scroll_line_start=moveto;
		text->vadj->value=(gfloat)moveto;
		gtk_adjustment_changed(text->vadj);
	}

}
static LineCache * goto_line(GtkExText *text,gint posindex,gboolean select,gboolean killselect)
{
	gint curpos,oldpos,len;
	LineCache *pos;

	oldpos=GTK_EDITABLE(text)->current_pos;
	curpos=text->line_cursor_index;
	pos=line_set(text,posindex);
	len=line_get_length(text,pos);
	if(curpos>len)	
		line_set_cursor_pos(text,len-1);
	else
		line_set_cursor_pos(text,curpos);

	scroll_to_view(text,text->line_pos_index);
	update_select(text,oldpos,select,killselect);
	
	return text->line_pos;

}

gint move_cursor(GtkExText *text,gint moveto,gboolean select)
{
	/* if move cursor we should always kill selection if select is false */ 
	gint oldpos;

	oldpos=GTK_EDITABLE(text)->current_pos;
	line_set_cursor_pos(text, moveto);

	scroll_to_view(text,text->line_pos_index);

	update_select(text,oldpos,select,TRUE);	

	return GTK_EDITABLE(text)->current_pos;
}
static void update_select(GtkExText *text,gint oldpos,gboolean select,gboolean killselect)
{
	if(select)	{ 
		/* FORWARD SELECT */
		if(oldpos<GTK_EDITABLE(text)->current_pos)
		{
			if(GTK_EDITABLE(text)->selection_start_pos==-1  )	{	
					GTK_EDITABLE(text)->selection_start_pos=oldpos;
					GTK_EDITABLE(text)->selection_end_pos=GTK_EDITABLE(text)->current_pos;	
			}
			else if(GTK_EDITABLE(text)->selection_start_pos==oldpos)
					GTK_EDITABLE(text)->selection_start_pos=GTK_EDITABLE(text)->current_pos;
			else 
					GTK_EDITABLE(text)->selection_end_pos=GTK_EDITABLE(text)->current_pos;	
		}
		/*  BACKWARD SELECT */
		else 
		{
			if(GTK_EDITABLE(text)->selection_end_pos==-1)	{
				GTK_EDITABLE(text)->selection_end_pos=oldpos;
				GTK_EDITABLE(text)->selection_start_pos=GTK_EDITABLE(text)->current_pos;
			}
			else if(GTK_EDITABLE(text)->selection_end_pos==oldpos)
					GTK_EDITABLE(text)->selection_end_pos=GTK_EDITABLE(text)->current_pos;
			else 
					GTK_EDITABLE(text)->selection_start_pos=GTK_EDITABLE(text)->current_pos;	
		}
		/* this is to make sure start end is not swaped */
		if(GTK_EDITABLE(text)->selection_end_pos<GTK_EDITABLE(text)->selection_start_pos)
		{
			gint swap=GTK_EDITABLE(text)->selection_start_pos;
			GTK_EDITABLE(text)->selection_start_pos=GTK_EDITABLE(text)->selection_end_pos;
			GTK_EDITABLE(text)->selection_end_pos=swap;
		}
	}	
	else	if(killselect){	
		GTK_EDITABLE(text)->selection_start_pos=-1;
		GTK_EDITABLE(text)->selection_end_pos=-1;
	}
}
static gint gtk_extext_expose(GtkWidget *widget,GdkEventExpose *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  
  if(event->window==GTK_EXTEXT(widget)->text_area)	
	expose_text(GTK_EXTEXT(widget),&event->area ,TRUE);
  else if(event->count == 0)
  {
      gtk_widget_draw_focus (widget);
  }
	return FALSE;
}

GtkType 
gtk_extext_get_type(void)
{
	static GtkType extext_type=0;

	if(!extext_type)
	{
		static const GtkTypeInfo extext_info=	
		{	
			"GtkExText",
			sizeof(GtkExText),
			sizeof(GtkExTextClass),
			(GtkClassInitFunc) gtk_extext_class_init,
			(GtkObjectInitFunc) gtk_extext_init,
			NULL,
			NULL,
			(GtkClassInitFunc)NULL,
		};
		extext_type=gtk_type_unique(GTK_TYPE_EDITABLE,&extext_info);
	}
	return extext_type;
}

/*****************************************************************************/
/*                               Public USER API                                 */
/*****************************************************************************/

void       gtk_extext_insert          (GtkExText       *text,
				     GdkFont       *font,
				     GdkColor      *fore,
				     GdkColor      *back,
				     const char    *chars,
				     gint           length)
{
	gint pos=GTK_EDITABLE(text)->current_pos;
	if(length==-1)
		length=strlen(chars);

	g_warning("gtk_extext_insert is OBSOLETE use gtk_editable_insert_text\n"
					"or gtk_extext_insert_with_style_and_property\n");

	gtk_editable_insert_text(GTK_EDITABLE(text),chars,length,&pos);
}

gpointer gtk_extext_line_get_data_and_pos(GtkExText *text,gint line,
					gint *startpos,gint *endpos)
{
	LineCache *cache=NULL;
	gint lpos=0;
  g_return_val_if_fail (text != NULL, NULL);
  g_return_val_if_fail (GTK_IS_EXTEXT (text), NULL);

	if(!startpos || !endpos)
		return NULL;

	*startpos=0;
	*endpos=0;
	while(cache)	{
		*endpos=*endpos + cache->len;
		if(lpos==line)
			return cache->data;

		lpos++;
		*startpos=*startpos + cache->len;
		cache=cache->next;
	};	
	
	return NULL;
}

gpointer gtk_extext_line_get_data(GtkExText *text,gint line)
{
	gint s,e;
	LineCache *cache=NULL;

  g_return_val_if_fail (text != NULL, NULL);
  g_return_val_if_fail (GTK_IS_EXTEXT (text), NULL);

	cache=gtk_extext_line_get_data_and_pos(text,line,&s,&e);

	return cache ? cache->data : NULL; 
}

void gtk_extext_line_set_data(GtkExText *text,gint line,gpointer data)
{
	LineCache *cache;
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));

	cache=line_get_ptr(text,&line);	
	if(cache) 
		 cache->data=data;
}

gint gtk_extext_set_line(GtkExText *text,gint line)
{
	line=line_get_num(text,goto_line(text,line,FALSE,TRUE),0);
	gtk_widget_queue_draw(GTK_WIDGET(text));
	return line;
}

gint gtk_extext_set_column(GtkExText *text,gint col)
{
	col= line_set_cursor_pos(text,col);
	gtk_widget_queue_draw(GTK_WIDGET(text));
	return col;
}

gint gtk_extext_get_line(GtkExText *text)
{
	return text->line_pos_index;
}

gint gtk_extext_get_column(GtkExText *text)
{
	return text->line_cursor_index;
}

void gtk_extext_set_editable(GtkEditable *edit,gboolean setit)
{
	edit->editable=setit;
	gtk_widget_queue_draw(GTK_WIDGET(edit));
}
void gtk_extext_freeze(GtkExText *text)
{
	text->freeze=TRUE;	
}
void gtk_extext_thaw(GtkExText *text)
{
	text->freeze=FALSE;

	goto_line(text,text->line_pos_index,FALSE,TRUE);
	
	gtk_widget_queue_draw(GTK_WIDGET(text));
}
static void gtk_extext_update_text(GtkEditable *text,gint start,gint end)
{
	/* hmm not sure about this */
	gtk_widget_queue_draw(GTK_WIDGET(text));
}
void gtk_extext_undo_set_max(GtkExText *text,gint max)
{
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));

  text->undo_max=max;
}

gboolean gtk_extext_undo_is_emty(GtkExText *text)
{
  g_return_val_if_fail (text != NULL,FALSE);
  g_return_val_if_fail (GTK_IS_EXTEXT (text),FALSE );

  return text->undo_buffert ? TRUE : FALSE ;
}

gboolean gtk_extext_redo_is_emty(GtkExText *text)
{
  g_return_val_if_fail (text != NULL,FALSE );
  g_return_val_if_fail (GTK_IS_EXTEXT (text), FALSE);

  return text->undo_buffert->next ? TRUE : FALSE ;
}

void gtk_extext_undo_clear_all(GtkExText *text)
{
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));
  
  g_slist_foreach(text->undo_buffert,extext_undo_foreach,NULL);
  g_slist_free(text->undo_buffert);
  text->undo_buffert=NULL;

  if (GTK_WIDGET_REALIZED (text))
	  gtk_signal_emit(GTK_OBJECT(text),extext_signals[UNDO_CHANGED]);
}

void gtk_extext_undo(GtkExText *text)
{
  GSList *cur;
  UndoBuffert *undo;
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));
  if(!text->undo_buffert)	
	return ;

  text->undo_flag=TRUE;
  cur=g_slist_last(text->undo_buffert);    
  undo=(UndoBuffert *)cur->data;	
  switch(undo->type) {
	case EXTEXT_UNDO_INSERT:
		gtk_editable_delete_text(GTK_EDITABLE(text),undo->startpos,undo->endpos);
	break;
	case EXTEXT_UNDO_REMOVE:
		gtk_editable_insert_text(GTK_EDITABLE(text), undo->buffert,
												undo->endpos - undo->startpos,&undo->startpos);
	break;
  };
 
  text->undo_flag=FALSE;

  g_free(undo);
  text->undo_buffert=g_slist_remove_link(text->undo_buffert,cur); 
  text->undo_last--;
  text->undo_count--;
/* the last element */
  if(!text->undo_buffert)  
	gtk_signal_emit(GTK_OBJECT(text),extext_signals[UNDO_CHANGED]);
}

void gtk_extext_redo(GtkExText *text)
{
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_EXTEXT (text));
	
}


/*****************************************************************************/
/*           PUBLIC                  editable overrides                               */
/*****************************************************************************/

gint gtk_extext_line_foreach(GtkExText *text,gint lindex,LineForEachFunc func,gpointer userdata)
{
	LineCache *cur;
	gint start,i;
	start=0;
	i=0;
	
	if(lindex>text->line_count)
		return -1;	

	cur=text->line_start;
	while(cur)
	{
		if(i>=lindex)	{
			if(!(*func)(text,i,start,start+ cur->len,cur->data,userdata))
				return i+1;
		}
		i++;
		start+=cur->len;
		cur=cur->next;
	};
	return -1;
}

void gtk_extext_insert_text       (GtkEditable       *editable,
					const gchar       *new_text,
					gint               new_len,
					gint               *position)
{
	GtkExText *text;
  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_EDITABLE (editable));
  g_return_if_fail (GTK_IS_EXTEXT (editable));

	text=GTK_EXTEXT(editable);
	
	if(GTK_EDITABLE(text)->selection_start_pos!=-1 && GTK_EDITABLE(text)->selection_end_pos!=-1)	{
		if(*position > editable->selection_start_pos  && *position<=editable->selection_end_pos) {
			gtk_editable_delete_text(editable,editable->selection_start_pos,editable->selection_end_pos);
			*position=editable->current_pos;
		}
		else
			gtk_editable_delete_text(editable,editable->selection_start_pos,editable->selection_end_pos);

	}

	*position=text_insert(text,*position,(gchar*)new_text,new_len,TRUE);
	
	text->vadj->upper=text->line_count;
	if(text->scroll_line_count)
		text->vadj->page_size=text->line_count/text->scroll_line_count;

	gtk_adjustment_changed(text->vadj);
	gtk_widget_queue_draw(GTK_WIDGET(text));
}

static void gtk_extext_delete_text       (GtkEditable       *editable,
					gint              startpos,
					gint               endpos)
{
	GtkExText  *text;
	gint oldlinecount;
  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_EDITABLE (editable));
  g_return_if_fail (GTK_IS_EXTEXT (editable));

	text=GTK_EXTEXT(editable);	
	oldlinecount=GTK_EXTEXT(editable)->line_count;	
	if(editable->selection_start_pos!=editable->selection_end_pos && editable->selection_start_pos!=-1)
		text_remove(GTK_EXTEXT(editable),editable->selection_start_pos,
						 editable->selection_end_pos - editable->selection_start_pos);
	else
		text_remove(GTK_EXTEXT(editable),startpos,endpos - startpos);

	if(oldlinecount!=text->line_count)	{
		text->vadj->upper=text->line_count;
		if(text->scroll_line_count)
			text->vadj->page_size=text->line_count/text->scroll_line_count;

		gtk_adjustment_changed(text->vadj);
	}
	gtk_widget_queue_draw(GTK_WIDGET(editable));
}
static void gtk_extext_select_region(GtkEditable *editable,gint start,gint end)
{
	editable->selection_start_pos=start;
	editable->selection_end_pos=end;

	gtk_widget_queue_draw(GTK_WIDGET(editable));
}
static gchar *gtk_extext_get_chars(GtkEditable *editable,gint start,gint end)
{
	return text_get_text(GTK_EXTEXT(editable),start,end);
}

void gtk_extext_set_position(GtkEditable *editable,gint pos)
{
	text_set_pos(GTK_EXTEXT(editable),pos);
	goto_line(GTK_EXTEXT(editable),GTK_EXTEXT(editable)->line_pos_index,FALSE,TRUE);

	gtk_widget_queue_draw(GTK_WIDGET(editable));
}

gint gtk_extext_get_position(GtkEditable *edit)
{
	return GTK_EDITABLE(edit)->current_pos;
}

gint gtk_extext_get_length(GtkExText *text)
{
   g_return_val_if_fail (GTK_IS_EXTEXT (text), FALSE);
 
	return text->length;	
}

#ifdef WITH_REGEX
/*****************************************************************************/
/*                                  Regex API                                    */
/* 	stolen from GtkEditor by Thomas Maulund */
/*****************************************************************************/

/* regex_search -- searches for regex in text from position
 * 'start'. If 'forward' is TRUE, it searches forward, otherwise
 * backwards.  If  found, returns index of beginning of the match,
 * otherwise returns < 0.  If found the match interval is
 * [m.from,m.to[.  m.from is always equal to the return value, but
 * m.to is undefined if no match.  I search in GtkSCText, since I don't
 * really need an editor for this function. This will also make it
 * easier to move this function to the text widget, should we want
 * that. */
gint gtk_extext_regex_search (GtkExText *text, gint start, Regex *regex,
	      gboolean forward, Match *m)
{
  g_return_val_if_fail (start <= text->length, -1);
  g_return_val_if_fail (regex != NULL, -1);
  g_return_val_if_fail (m != NULL, -1);

  m->from = re_search_2 (&regex->buf,
			 /* text before gap */
			 text->text,
			text->part1len,
			 /* text after gap */
			 &text->part2text[text->part1len+1],
				text->length - text->part1len,
			 /* from 'start' and forward to the end */
			 start,
			 (forward ? text->length - start
			  : -start),
			 &regex->reg,
			 text->length - start);

  if (m->from > -1) m->to = regex->reg.end[0];

  return m->from;    
}

/* regex_match -- tries to match regex at the 'pos' position in the
 * text. It returns the number of chars matched, or -1 if no match.
 * Warning!  The number matched can be 0, if the regex matches the
 * empty string.  The reason for workin on GtkSCText is the same as in
 * regex_search. */
gint gtk_extext_regex_match (GtkExText *text, gint pos, Regex *regex)
{
  g_return_val_if_fail (pos <= text->length, -1);
  g_return_val_if_fail (regex != NULL, -1);

  return re_match_2 (&regex->buf,
		     /* text before gap */
		     text->text, text->part1len,
		     /* text after gap */
		     &text->part2text[text->part1len+1],
		     text->length - text->part1len,
		     /* from pos and not after the end */
		     pos, &regex->reg, text->length);
}
#endif
