/*** jdk.c ***/

#include "jdk.h"

/*
    This function is responsible for compiling java source files
    to class files.  It uses a fork call to execute the jdk
    compiler - javac, and passes the full path to the file it
    wishes to have compiled to javac as an argument.  Files must
    be saved in order to be compiled.
*/
void	jdk_compile(GtkWidget *widget, struct cmj *codemonkey)
{
	gint stdin_pipe[2];
	gint stdout_pipe[2];
	gint stderr_pipe[2];
	
	gchar buffer[BUFSIZ+1];
	gchar buffer_err[BUFSIZ+1];
	gint fork_result;
	gint data_processed;
	gint data_processed_err;

	guint nchars=0;
	GdkColormap	*cmap;
	GdkColor	color;

	if(codemonkey->current == NULL)
	{
		return;
	}

	if((codemonkey->current->disk_copy)==0)
	{
		return;
	}

	memset(buffer, '\0', sizeof(buffer));
	memset(buffer_err, '\0', sizeof(buffer));

	nchars = gtk_text_get_length(GTK_TEXT(codemonkey->current->jdk_box));
	gtk_text_set_point(GTK_TEXT(codemonkey->current->jdk_box), (guint)0);
	gtk_text_forward_delete(GTK_TEXT(codemonkey->current->jdk_box), nchars);

	if( (pipe(stdin_pipe) == 0)
		&& (pipe(stdout_pipe) == 0)
		&& (pipe(stderr_pipe) == 0)
	  )
	{
		fork_result = fork();
		if(fork_result == -1)
		{
			fprintf(stderr, "Fork Failure\n");
			exit(EXIT_FAILURE);
		}
		else if(fork_result == 0)
		{
			close(0);
			dup(stdin_pipe[0]);
			close(stdin_pipe[0]);
			close(stdin_pipe[1]);

			close(1);
			dup(stdout_pipe[1]);
			close(stdout_pipe[0]);
			close(stdout_pipe[1]);

			close(2);
			dup(stderr_pipe[1]);
			close(stderr_pipe[0]);
			close(stderr_pipe[1]);
			
			execlp("javac","javac", "-d", codemonkey->current->class_path, codemonkey->current->file_name, (char *)0);

			exit(EXIT_FAILURE);
		}
		else
		{
			cmap = gdk_colormap_get_system();
			color.red = 0xffffff;
			color.green = 0;
			color.blue = 0;
			if(!gdk_color_alloc(cmap,&color))
			{
				g_error("Couldn't allocate color");
			}
			close(stdin_pipe[0]);
			close(stdin_pipe[1]);
			close(stderr_pipe[1]);
			gtk_text_freeze(GTK_TEXT(codemonkey->current->jdk_box));
			while(1)
			{
				data_processed_err=read(stderr_pipe[0],buffer_err,BUFSIZ);
				gtk_text_insert(GTK_TEXT(codemonkey->current->jdk_box),NULL,&color,NULL,buffer_err,data_processed_err);
				if(data_processed_err == 0) break;
			}
			close(stderr_pipe[0]);
			close(stdout_pipe[1]);
			while(1)
			{
				data_processed=read(stdout_pipe[0],buffer,BUFSIZ);
				gtk_text_insert(GTK_TEXT(codemonkey->current->jdk_box),NULL,NULL,NULL,buffer, data_processed);
				if(data_processed == 0) break;
			}
			close(stdout_pipe[0]);
			gtk_text_thaw(GTK_TEXT(codemonkey->current->jdk_box));
		}
	}
	return;
}

/*
    This function is responsible for launching the java runtime
    to execute a class file.  This function works by using a 
    fork call to launch the java runtime.  It passes the location
    of the class file it wishes to execute as an argument.
*/
void	jdk_execute(GtkWidget *widget, struct cmj *codemonkey)
{
	gint stdin_pipe[2];
	gint stdout_pipe[2];
	gint stderr_pipe[2];
	gchar buffer[BUFSIZ+1];
	gchar buffer_err[BUFSIZ+1];
	gint fork_result;
	gint data_processed;
	gint data_processed_err;

	guint nchars = 0;
	GdkColormap	*cmap;
	GdkColor	color;

	if(codemonkey->current == NULL)
	{
		return;
	}

	if((codemonkey->current->disk_copy)==0)
	{
		return;
	}

	memset(buffer, '\0', sizeof(buffer));
	memset(buffer_err, '\0', sizeof(buffer_err));

	nchars = gtk_text_get_length(GTK_TEXT(codemonkey->current->jdk_box));
	gtk_text_set_point(GTK_TEXT(codemonkey->current->jdk_box), (guint)0);
	gtk_text_forward_delete(GTK_TEXT(codemonkey->current->jdk_box), nchars);

	if( (pipe(stdin_pipe) == 0)
		&& (pipe(stdout_pipe) == 0)
		&& (pipe(stderr_pipe) == 0)
	  )
	{
		fork_result = fork();
		if(fork_result == -1)
		{
			fprintf(stderr, "Fork Failure\n");
			exit(EXIT_FAILURE);
		}
		else if(fork_result == 0)
		{
			close(0);
			dup(stdin_pipe[0]);
			close(stdin_pipe[0]);
			close(stdin_pipe[1]);

			close(1);
			dup(stdout_pipe[1]);
			close(stdout_pipe[0]);
			close(stdout_pipe[1]);

			close(2);
			dup(stderr_pipe[1]);
			close(stdout_pipe[0]);
			close(stdout_pipe[1]);

			execlp("java","java","-cp", codemonkey->current->class_path, codemonkey->current->class_name, (char *)0);
			exit(EXIT_FAILURE);
		}
		else
		{
			cmap = gdk_colormap_get_system();
			color.red = 0xffffff;
			color.green = 0x000000;
			color.blue = 0x000000;
			if(!gdk_color_alloc(cmap,&color))
			{
				g_error("Couldn't allocate color");
			}
			close(stdin_pipe[0]);
			close(stdin_pipe[1]);
			close(stderr_pipe[1]);
			gtk_text_freeze(GTK_TEXT(codemonkey->current->jdk_box));
			while(1)
			{
				data_processed_err=read(stderr_pipe[0],buffer_err,BUFSIZ);
				gtk_text_insert(GTK_TEXT(codemonkey->current->jdk_box),NULL,&color,NULL,buffer_err,data_processed_err);
				if(data_processed_err == 0) break;
			}
			close(stderr_pipe[0]);
			close(stdout_pipe[1]);

			while(1)
			{
				data_processed=read(stdout_pipe[0],buffer,BUFSIZ); 
			gtk_text_insert(GTK_TEXT(codemonkey->current->jdk_box),NULL,NULL,NULL,buffer,data_processed);
				if(data_processed == 0) break;
			}
			close(stdout_pipe[0]);

			gtk_text_thaw(GTK_TEXT(codemonkey->current->jdk_box));
		}
	}

	return;
}


/*
    This function clears the output in the output display.  It 
    clears all text and error messages from the display which
    is an instance of the GTK widget GTK_TEXT.
*/
void	jdk_clear(GtkWidget *widget, struct cmj *codemonkey)
{
	guint nchars=0;

	if(codemonkey->current == NULL)
	{
		return;
	}

	nchars = gtk_text_get_length(GTK_TEXT(codemonkey->current->jdk_box));
	gtk_text_set_point(GTK_TEXT(codemonkey->current->jdk_box), (guint)0);
	gtk_text_forward_delete(GTK_TEXT(codemonkey->current->jdk_box), nchars);

	return;
}
